@@ -8,6 +8,7 @@
  * 14 Nov 2000 ht - for VMS, truncated name of long functions to under 32 char
  */
 
+#define IN_LIBXML
 #include "libxml.h"
 
 #include <string.h>
@@ -2080,6 +2081,19 @@ xmlParserInputBufferPush(xmlParserInputBufferPtr in,
     return(nbchars);
 }
 
+/**
+ * endOfInput:
+ *
+ * When reading from an Input channel indicated end of file or error
+ * don't reread from it again.
+ */
+static int
+endOfInput (void * context ATTRIBUTE_UNUSED,
+	    char * buffer ATTRIBUTE_UNUSED,
+	    int len ATTRIBUTE_UNUSED) {
+    return(0);
+}
+
 /**
  * xmlParserInputBufferGrow:
  * @in:  a buffered parser input
@@ -2125,6 +2139,8 @@ xmlParserInputBufferGrow(xmlParserInputBufferPtr in, int len) {
      */
     if (in->readcallback != NULL) {
 	res = in->readcallback(in->context, &buffer[0], len);
+	if (res <= 0)
+	    in->readcallback = endOfInput;
     } else {
         xmlGenericError(xmlGenericErrorContext,
 		"xmlParserInputBufferGrow : no input !\n");
@@ -2528,6 +2544,8 @@ xmlDefaultExternalEntityLoader(const char *URL, const char *ID,
 	resource = (xmlChar *) URL;
 
     if (resource == NULL) {
+	if (ID == NULL)
+	    ID = "NULL";
 	if ((ctxt->validate) && (ctxt->sax != NULL) && 
             (ctxt->sax->error != NULL))
 	    ctxt->sax->error(ctxt,
